module beams
  use const_and_precisions, only : wp_
  implicit none

contains

  subroutine read_beam0(file_beam,alpha0,beta0,fghz,x00,y00,z00,              &
                        wcsi,weta,rcicsi,rcieta,phiw,phir,unit)
    use const_and_precisions, only : pi,vc=>ccgs_
    use utils, only : get_free_unit
    implicit none
! arguments
    character(len=*), intent(in) :: file_beam
    real(wp_), intent(in) :: alpha0,beta0
    real(wp_), intent(out) :: fGHz,x00,y00,z00
    real(wp_), intent(out) :: wcsi,weta,rcicsi,rcieta,phir,phiw
    integer, intent(in), optional :: unit
! local variables
    integer :: u
    real(wp_) :: ak0,zrcsi,zreta,w0csi,w0eta,d0csi,d0eta
   
    if (present(unit)) then
      u=unit
    else
      u = get_free_unit()
    end if
    open(unit=u,file=trim(file_beam),status='OLD',action='READ')

!     fghz      wave frequency (GHz)     
    read(u,*) fGHz
!     x00,y00,z00 coordinates of launching point in cm
    read(u,*) x00, y00, z00
!     beams parameters in local reference system
!     w0 -> waist, d0 -> waist distance from launching point
!     phiw  angle of spot ellipse
    read(u,*) w0csi,w0eta,d0csi,d0eta,phiw
    close(u)

    ak0=2.0e9_wp_*pi*fghz/vc
    zrcsi=0.5_wp_*ak0*w0csi**2
    zreta=0.5_wp_*ak0*w0eta**2
    
    wcsi=w0csi*sqrt(1.0_wp_+(d0csi/zrcsi)**2)
    weta=w0eta*sqrt(1.0_wp_+(d0eta/zreta)**2)
    rcicsi=-d0csi/(d0csi**2+zrcsi**2)
    rcieta=-d0eta/(d0eta**2+zreta**2)
    phir=phiw

  end subroutine read_beam0



  subroutine read_beam1(file_beam,alpha0,beta0,fghz,x00,y00,z00,            &
                        wcsi,weta,rcicsi,rcieta,phiw,phir,unit)
    use const_and_precisions, only : pi,vc=>ccgs_
    use simplespline, only : spli, difcs
    use utils, only : get_free_unit,locate
    implicit none
! arguments
    character(len=*), intent(in) :: file_beam
    real(wp_), intent(in) :: alpha0
    real(wp_), intent(out) :: fghz,x00,y00,z00,beta0
    real(wp_), intent(out) :: wcsi,weta,rcicsi,rcieta,phir,phiw
    integer, intent(in), optional :: unit
! local variables
    integer :: u,ierr,iopt,ier,nisteer,i,k,ii
    real(wp_) :: steer,alphast,betast,x00mm,y00mm,z00mm
    real(wp_) :: w1,w2,rci1,rci2,phi1,phi2,dal
    real(wp_), dimension(:), allocatable :: alphastv,betastv,x00v,y00v,     &
                           z00v,waist1v,waist2v,rci1v,rci2v,phi1v,phi2v,    &
                           cbeta,cx0,cy0,cz0,cwaist1,cwaist2,               &
                           crci1,crci2,cphi1,cphi2
    
    if (present(unit)) then
      u=unit
    else
      u = get_free_unit()
    end if
    open(unit=u,file=file_beam,status='OLD',action='READ')

    read(u,*) fghz
    
    read(u,*) nisteer

    allocate(alphastv(nisteer),betastv(nisteer),waist1v(nisteer),     &
             waist2v(nisteer),rci1v(nisteer),rci2v(nisteer),          &
             phi1v(nisteer),phi2v(nisteer),x00v(nisteer),             &
             y00v(nisteer),z00v(nisteer),cbeta(4*nisteer),            &
             cx0(4*nisteer),cy0(4*nisteer),cz0(4*nisteer),            &
             cwaist1(4*nisteer),cwaist2(4*nisteer),crci1(4*nisteer),  &
             crci2(4*nisteer),cphi1(4*nisteer),cphi2(4*nisteer),      &
             stat=ierr)
    if (ierr/=0) then
      close(u)
    deallocate(alphastv,betastv,waist1v,waist2v,rci1v,rci2v,          &
               phi1v,phi2v,x00v,y00v,z00v,cbeta,                 & 
               cx0,cy0,cz0,cwaist1,cwaist2,crci1,crci2,cphi1,cphi2)
      write(*,*) 'cannot allocate arrays for beam data'
      stop
    end if

    do i=1,nisteer
      read(u,*) steer,alphast,betast,x00mm,y00mm,z00mm,            &
                     w1,w2,rci1,rci2,phi1,phi2
! initial beam data measured in mm -> transformed to cm
      x00v(i)=0.1_wp_*x00mm
      y00v(i)=0.1_wp_*y00mm
      z00v(i)=0.1_wp_*z00mm
      alphastv(i)=alphast
      betastv(i)=betast
      waist1v(i)=0.1_wp_*w1
      rci1v(i)=1.0e1_wp_*rci1
      waist2v(i)=0.1_wp_*w2
      rci2v(i)=1.0e1_wp_*rci2
      phi1v(i)=phi1
      phi2v(i)=phi2
    end do
    close(u)
  
    iopt=0
    call difcs(alphastv,betastv,nisteer,iopt,cbeta,ier)
    call difcs(alphastv,waist1v,nisteer,iopt,cwaist1,ier)
    call difcs(alphastv,rci1v,nisteer,iopt,crci1,ier)
    call difcs(alphastv,waist2v,nisteer,iopt,cwaist2,ier)
    call difcs(alphastv,rci2v,nisteer,iopt,crci2,ier)
    call difcs(alphastv,phi1v,nisteer,iopt,cphi1,ier)
    call difcs(alphastv,phi2v,nisteer,iopt,cphi2,ier)
    call difcs(alphastv,x00v,nisteer,iopt,cx0,ier)
    call difcs(alphastv,y00v,nisteer,iopt,cy0,ier)
    call difcs(alphastv,z00v,nisteer,iopt,cz0,ier)

    if((alpha0 > alphastv(1)).and.(alpha0 < alphastv(nisteer))) then
      call locate(alphastv,nisteer,alpha0,k)
      dal=alpha0-alphastv(k)
      beta0=spli(cbeta,nisteer,k,dal)
      x00=spli(cx0,nisteer,k,dal)
      y00=spli(cy0,nisteer,k,dal)
      z00=spli(cz0,nisteer,k,dal)
      wcsi=spli(cwaist1,nisteer,k,dal)
      weta=spli(cwaist2,nisteer,k,dal)
      rcicsi=spli(crci1,nisteer,k,dal)
      rcieta=spli(crci2,nisteer,k,dal)
      phiw=spli(cphi1,nisteer,k,dal)
      phir=spli(cphi2,nisteer,k,dal)
    else
      write(*,*) '    alpha0 outside table range !!!'
      if(alpha0 >= alphastv(nisteer)) ii=nisteer
      if(alpha0 <= alphastv(1)) ii=1
      beta0=betastv(ii)
      x00=x00v(ii)
      y00=y00v(ii)
      z00=z00v(ii)
      wcsi=waist1v(ii)
      weta=waist2v(ii)
      rcicsi=rci1v(ii)
      rcieta=rci2v(ii)
      phiw=phi1v(ii)
      phir=phi2v(ii)
    end if

    deallocate(alphastv,betastv,waist1v,waist2v,rci1v,rci2v,          &
               phi1v,phi2v,x00v,y00v,z00v,cbeta,                 & 
               cx0,cy0,cz0,cwaist1,cwaist2,crci1,crci2,cphi1,cphi2)

  end subroutine read_beam1

  subroutine launchangles2n(alpha,beta,xv,anv)
    use const_and_precisions, only : degree
    implicit none
! arguments
    real(wp_), intent(in) :: alpha,beta,xv(3)
    real(wp_), intent(out) :: anv(3)
! local variables
    real(wp_) :: r,anr,anphi,a,b

    r=sqrt(xv(1)**2+xv(2)**2)
!    phi=atan2(y,x)
    print'(a,2f8.3)','alpha0, beta0 = ',alpha,beta
    print'(a,4f8.3)','x00, y00, R00, z00 = ',xv(1:2),r,xv(3)
    print*,' '
    a = degree*alpha
    b = degree*beta
!
!     angles alpha, beta in a local reference system as proposed by Gribov et al
!
    anr   = -cos(b)*cos(a)
    anphi =  sin(b)
!    anx   = -cos(b)*cos(a)
!    any   =  sin(b)
 
    anv(1)   = (anr*xv(1) - anphi*xv(2))/r  ! = anx
    anv(2)   = (anr*xv(2) + anphi*xv(1))/r  ! = any
!    anr   = (anx*xv(1) + any*xv(2))/r
!    anphi = (any*xv(1) - anx*xv(2))/r

    anv(3)   =-cos(b)*sin(a) ! = anz
  end subroutine launchangles2n

  subroutine xgygcoeff(fghz,ak0,bres,xgcn)
    use const_and_precisions, only : qe=>ecgs_,me=>mecgs_,vc=>ccgs_,pi,wce1_
    implicit none
! arguments
    real(wp_), intent(in) :: fghz
    real(wp_), intent(out) :: ak0,bres,xgcn
! local variables
    real(wp_) :: omega

    omega=2.0e9_wp_*pi*fghz     ! [rad/s]
    ak0=omega/vc                ! [rad/cm]
!
!     yg=btot/bres
!
    bres=omega/wce1_            ! [T]
!
!     xg=xgcn*dens19
!
    xgcn=4.0e13_wp_*pi*qe**2/(me*omega**2) ! [10^-19 m^3]
  end subroutine xgygcoeff
end module beams
